/* Invoke fcntl-like functions, but avoid some glitches.

   Copyright (C) 2005, 2009-2026 Free Software Foundation, Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */

/* Written by Paul Eggert.  */

#include <sys/types.h>

#ifdef __cplusplus
extern "C" {
#endif


int open_safer (char const *, int, ...);
int creat_safer (char const *, mode_t);

#if GNULIB_OPENAT_SAFER
int openat_safer (int, char const *, int, ...);
#endif


#ifdef __cplusplus
}
#endif
